/*
 * =====================================================================================
 *
 *       Filename:  depolarizer.cpp
 *
 *    Description:  Depolarizer implemenation
 *
 *        Version:  1.0
 *        Created:  27.05.2017 12:40:21
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *   Organization:  Budker Insitute of Nuclear Physics
 *
 * =====================================================================================
 */

#include <chrono>
#include <memory>
#include <optional>
#include <functional>
#include <fmt/core.h>
#include <fmt/format.h>
#include <fmt/chrono.h>
#include <fmt/ostream.h>
//#include <fmt/printf.h>
//#include <fmt/print.h>


#include "Depolarizer.h"
#include "camac/Generator.h"
#include "camac/Attenuator.h"


std::random_device RANDOM_DEVICE;
std::mt19937 RG(RANDOM_DEVICE());

using namespace std::chrono;

Depolarizer::Depolarizer(
		std::string name, 
		std::shared_ptr<device::Generator>  gen, 
		std::shared_ptr<device::Attenuator> att, 
		std::shared_ptr<device::DelayLine>   dl
		) 
	: generator(gen), attenuator(att), delay_line(dl)
{
	info.type  = "Depolarizer";
	info.model = "";
	info.address = "";
	info.name = name;
	isscan = false;
	ison = false;
	atomic_isscan.store(false,std::memory_order_relaxed );

	speed = 1; // Hz/s
	initial_frequency = 500000; //Hz
	harmonic_number = 3;
	attenuation = 55; //dB
	set_step_time();
	attenuator->SetAttenuation(attenuation);
	//last_frequency_list_point_iterator = flist.before_begin();
}

Depolarizer::~Depolarizer(void) {}

void Depolarizer::start_scan(void) 
{
	//if(!isscan) {
	if(not atomic_isscan.load(std::memory_order_relaxed)) {
		clear_frequency_list();
		begin_time = steady_clock::now();
		local_ifreq = initial_frequency;
		local_begin_time = begin_time;
		scan_counter++; //увеличим счётчик сканирований на 1
		start_thread();
	}
}

void Depolarizer::clear_frequency_list(void) {
		clear_barrier.init();
		const int number_of_scans_kept=10;
		if (scan_list.size() > number_of_scans_kept) {
			auto rit = scan_list.rbegin();
			std::advance(rit, number_of_scans_kept);
			auto t_cut = rit->first.timestamp;
			fmt::print("tcut = {:%F %T} \n", t_cut);
			//Найдем запись, где время больше tcut
			auto it = std::find_if(flist.begin(), flist.end(), 
					[&t_cut](const auto & state) { return state.timestamp>t_cut; } 
					);
			//if ( it != flist.end() && it != flist.before_begin() ){
			//	flist.erase_after(flist.before_begin(), it);
			//}
			if ( it != flist.end()){
				flist.erase(flist.begin(), it);
			}
			//for(const auto & s : flist) {
			//	fmt::print("{:%F %T}  {}\n", s.timestamp, s.frequency);
			//}
		}
		// remove all item except last one
		//if(last_frequency_list_point_iterator != flist.before_begin()) flist.erase_after(flist.before_begin(), last_frequency_list_point_iterator);
		clear_barrier.release();
}

void Depolarizer::continue_scan(void) 
{
	//if(!isscan) {
	if(not atomic_isscan.load(std::memory_order_relaxed)) {
		local_ifreq = last_frequency;
		local_begin_time = steady_clock::now();
		start_thread();
	}
}

void Depolarizer::start_thread(void) 
{
	//isscan = true;
	atomic_isscan.store(true,std::memory_order_relaxed );
	try {
		if(!scan_thread.joinable()) {
			scan_thread = std::thread([this]{scan();});
		}
	} 
	catch(std::runtime_error & e) {
		fmt::print("ERROR: {}\n", e.what());
	}
}

void Depolarizer::stop_scan(void) 
{
	if(scan_thread.joinable()) {
		atomic_isscan.store(false,std::memory_order_relaxed );
		scan_thread.join();
	}
}

void Depolarizer::pause_scan(void) 
{
	stop_scan();
}

//struct Row {
//	std::string name;
//	size_t width;
//	std::string title;
//	std::string format;
//};
//
//std::vector<Row> Table = 
//{
//	{"time", 23, "time", ""},
//	{"et", 10, "elapsed", ".3"},
//	{"freq", 13, "f", ".3"},
//	{"E", 13, "E", ".3"}
//};
//
//auto print_head2 = [] () {
//	std::string line;
//	for(auto & item : Table) {
//		line += fmt::format("{"+
//	}
//
//	fmt::print("#{:-^{}}\n","",line.size());
//	fmt::print("{}",line);
//	fmt::print("#{:-^{}}\n","", line.size());
//};


constexpr const char * FORMAT_STRING = "{time:^22}" 
                            "{et:>10}" 
							"{freq:>15}"
							"{E:>15}"     
							"{att:>7}"    
							"{speed_Hzs:>15}"
                            "{speed_kevs:>15}"       
                            "{step:>12}"
							"{step_time:>10}"
                            "{dynamic_width:>11}"
                            "{pdf_width:>10}"
                            "{total_width:>10}";

auto print_head = [] () {
	auto line = fmt::format( FORMAT_STRING, 
			fmt::arg("time",         "time"), 
			fmt::arg("et",           "elapsed") , 
			fmt::arg("freq",         "f,Hz"), 
			fmt::arg("E",            "E,MeV"), 
			fmt::arg("att",          "att"), 
			fmt::arg("speed_Hzs",    "Hz/s"), 
			fmt::arg("speed_kevs",   "keV/s"), 
			fmt::arg("step",         "df,Hz"), 
			fmt::arg("step_time",    "dt,s"), 
			fmt::arg("dynamic_width","dyn" ),
			fmt::arg("pdf_width",    "pdf"   ),
			fmt::arg("total_width",  "total" )
			);
	fmt::print("#{:-<{}}\n","",line.size()-1); //minus one  for take into account '#'
	fmt::print("{}\n",line);
	fmt::print("#{:-<{}}\n","", line.size()-1);
};

auto print_status = [] (const Depolarizer & d, std::chrono::steady_clock::duration  et) 
{
	using namespace std::chrono;
	auto t = system_clock::now();
//fmt::print("{time:%F %T}.{ms:03} {et:>13.3} {freq:>13.3f} Hz {E:>13.6f} MeV {att:>4.0f} dB {speed_Hzs:>8.3f} Hz/s {speed_kevs:>8.3f} keV/s {step:>7.3f} {step_time:>7.3} {dynamic_width:>6.3f} Hz {pdf_width:6.3f} Hz {total_width:6.3f} Hz\n", 
	fmt::print(FORMAT_STRING, 
			fmt::arg("time",          fmt::format("{:%F %T}.{:03}", t, (round<milliseconds>(t).time_since_epoch() % 1s).count()) ), 
			fmt::arg("et",            fmt::format("{:>.3}",  duration_cast<duration<double>>(et))), 
			fmt::arg("freq",          fmt::format("{:>.3f} Hz", d.get_frequency())), 
			fmt::arg("E",             fmt::format("{:>.3f} MeV", d.energy())), 
			fmt::arg("att",           fmt::format("{:>}dB", d.get_attenuation())), 
			fmt::arg("speed_Hzs",     fmt::format("{:>.3f} Hz/s", d.get_speed())), 
			fmt::arg("speed_kevs",    fmt::format("{:>.3f} keV/s", d.get_speed()*1e3/d.get_revolution()*DepolarizerState::RDM)), 
			fmt::arg("step",          fmt::format("{:>.3f} Hz", d.get_step())), 
			fmt::arg("step_time",     fmt::format("{:>.3}",  std::chrono::duration_cast< std::chrono::duration<double> >(d.get_step_time()))), 
			fmt::arg("dynamic_width", fmt::format("{:>.3f} Hz", sqrt( std::abs(d.get_speed() ) ) )),
			fmt::arg("pdf_width",     fmt::format("{:>.3f} Hz", d.get_width()   )),
			fmt::arg("total_width",   fmt::format("{:>.3f} Hz", sqrt( d.get_width()*d.get_width() + std::abs(d.get_speed() ) ) ))
			);
	fmt::print("\n");
};

void Depolarizer::add_state_to_change_list(bool is_scan_flag) {
	//is_scan_flag - модификатор, который говорит в каком состоянии был деполяризатор. Это нужно
	//чтобы не городить сложную логику регистрации состояния
	auto state = get_state();
	state.isscan = is_scan_flag;
	state_change_list.push_back(state);
}

void Depolarizer::scan(void) 
{
	DepolarizerState begin_state;
	try {
		isscan=true;
		add_state_to_change_list(false); //здесь мы регистрируем состояние до сканирования
		set_frequency(0.0);
		on();
		unsigned long count=0; //просто счётчик для печати заголовков
		while(atomic_isscan.load(std::memory_order_relaxed)) {
			auto t = steady_clock::now(); //begin time of the step
			std::chrono::duration<double> elapsed_time = t-local_begin_time;
			//прервем сканирование, если оно происходит больше 3х часов. В случае, если программа упала а мы забыли выключить деполяризатор
			if ( elapsed_time.count() > 3600*3 ) break;
			frequency = local_ifreq + speed*elapsed_time.count();

			if(pdf)  {
				double df = (*pdf)(RG);
				frequency +=  df;
			}

			set_frequency(frequency);

			if (count == 0) {//зарегистрируем состояние деполяризатора сразу после включения
				begin_state = get_state();
				add_state_to_change_list(true);
			}

			if ( verbose_level >= VerboseLevel::INFO ) {
				if(count%64==5) print_head();
				print_status(*this, steady_clock::now()-begin_time);
				std::cout << std::flush;
			}

			if ( step_time > 100ms ) { //if step time is large then
				while( atomic_isscan.load(std::memory_order_relaxed) && std::chrono::steady_clock::now()<t+step_time) {
					std::this_thread::sleep_for(10ms);
				}

			} else {
				std::this_thread::sleep_for(step_time);
			}

			++count;
		}
	}
	catch(std::runtime_error & e) {
		fmt::print(std::cerr, "Depolarizer::scan: ERROR in scan procedure: {}. Stop scan\n", e.what());
	}
	//isscan = false;
	last_frequency = frequency;
	add_state_to_change_list(true); //состояние непосредственно перед выключением
	DepolarizerState end_state = get_state();
	end_state.isscan=true;
	scan_list.emplace_back(begin_state, end_state);
	off();
	add_state_to_change_list(false); //а здесь сразу после выключения
	isscan=false;

	//for ( auto & s : state_change_list ) {
	//	fmt::print("{:%F %T} {:10} {:10} {:10} {:10.3f}\n", s.timestamp, s.isscan, s.speed, s.attenuation, s.frequency);
	//}

	//fmt::print("Scan list\n");
	//int n=0;
	//for (auto & scan : scan_list)  {
	//	auto & begin = scan.first;
	//	auto & end = scan.second;
	//	fmt::print("{:5} {:+10.3f} Hz/s ({:5.1f} dB) from {:%F %T} -> {:%F %T} {:>15.3f} Hz -> {:<15.3f} Hz\n", n, begin.speed, begin.attenuation, begin.timestamp, end.timestamp, begin.frequency, end.frequency);
	//	n++;
	//}
}



bool Depolarizer::is_on(void) const 
{ 
	return generator->IsOn(); 
}

void Depolarizer::on(void) 
{ 
	ison = true;
	generator->On(); 
}

void Depolarizer::on(double f) 
{ 
	generator->On(); 
	ison = true;
	set_frequency(f);
}

void Depolarizer::off(void) 
{ 
	generator->Off(); 
	ison = false;
	set_frequency(0);
}

void Depolarizer::set_frequency(double f)
{
	frequency = f;
	generator->SetFrequency(frequency);
	flist.emplace_back( 
			std::chrono::system_clock::now(), 
			isscan, 
			isscan,
			harmonic_number,
			frequency, 
			speed, 
			step,
			width,
			attenuation
			);
	//fmt::print("f = {}, speed = {}, attenuation = {}\n", frequency, speed, attenuation);
}

void Depolarizer::set_attenuation(double a)
{
	attenuation = a;
	if(attenuator) {
		attenuator->SetAttenuation(a);
	}
}

void Depolarizer::set_attenuation(void)
{
	if(attenuator)
	{
		attenuator->SetAttenuation(attenuation);
	}
}

double Depolarizer::get_attenuation(void) const 
{
	if(attenuator)
	{
		return attenuator->GetAttenuation();
	}
	else
	{
		return attenuation;
	}
}

void Depolarizer::set_step_time(void)
{
	using namespace std::chrono;
	constexpr double epsilon = std::numeric_limits<double>::epsilon();
	const double fdot = std::abs(speed);
	//system_clock::duration max_step_time = 1s;

	step_time = [&] () -> system_clock::duration {
		if ( width > epsilon)  return 10ms; //generate frequency with maximum rate if probability distribution is defined
		if ( fdot < epsilon )  return 1s;   //if scan speed is zero then change frequency every second
		auto dt = duration<double>(1.0/( std::sqrt(fdot)*STEP_NUMBER_IN_WIDTH)); //calculate step time
		if ( dt > 1s) return 1s; // defince maximum step time
		return duration_cast<system_clock::duration> (dt);
	}();

	//calculate step from step_time. 
	//NOTE: step is not used in scan algorithm. only step_time and elapsed time are used 
	//to determine current frequency to generate
	step =  fdot*duration_cast<duration<double>>(step_time).count();
}


void Depolarizer::set_speed(double s) {
	if(fabs(speed-s) > std::numeric_limits<double>::epsilon()) {
		local_ifreq = frequency;
		local_begin_time = steady_clock::now();
		speed=s;
		set_step_time();
	}
}

void Depolarizer::set_step(double)
{
	//step = s;
	//set_step_time();
}

void Depolarizer::set_initial(double f) 
{
	initial_frequency = f;
}

void Depolarizer::set_final(double f) 
{
	final_frequency = f;
}

void Depolarizer::set_width(double w=0) {
	//std::cout << "Depolarizer::set_width " << std::endl;
	width = std::abs(w); //width mast be non negative value
	if(width> std::numeric_limits<double>::epsilon()) {
		if constexpr (  std::is_same<  pdf_t, std::normal_distribution<> >::value ) {
			constexpr double FWHM_SIGMA = 2.35482004503094938200; // 2.0*sqrt(2.0*log(2.0));
			pdf.reset(new pdf_t(0,width/FWHM_SIGMA));
			//std::cout << "Set gauss distribution with width: " << width << " Hz" << std::endl;
		} else 
		if constexpr (  std::is_same<  pdf_t, std::uniform_real_distribution<> > :: value) {
			pdf.reset(new std::uniform_real_distribution<>( - 0.5*width, + 0.5*width));
			//std::cout << "Set uniform distribution with width: " << width << " Hz" << std::endl;
		}
	} else {
			//std::cout << "Disable pdf width = " << width << " Hz" << std::endl;;
		pdf.reset(nullptr);
	}
	set_step_time();
}

void Depolarizer::set_delay(double delay_ns) 
{
	std::chrono::nanoseconds d(static_cast<int>(delay_ns));
	delay_line->SetDelay(d);
}

double Depolarizer::get_delay(void) const 
{
	return static_cast<double>(delay_line->GetDelay().count());
}

//Linear interpolation between two point
//Обобщенная функция линейной интерполяции для 
//произвольного объекта
template<typename ProjY, typename ProjX, typename Type, typename TypeX> 
auto linear_interpol(
		const ProjY & proj_y, //функция извлекает значение y из объекта: y = proj_y(left)
		const ProjX & proj_x,  //функция ивзлекает значение x (типа TypeX) из объекта, x = proj_x(right)
		const Type & left, //первый объект
		const Type & right, //второй объект
		const TypeX & x //значение в этой точке надо интерполировать
		) 
-> typename std::remove_cv< typename std::remove_reference<decltype(std::invoke(proj_y, left))>::type >::type
//decltype(std::invoke(proj_y,left)) - тип значения y, но на всякий случай удалил из него константность и все ссылки
{
	//Вот ниже, собственно я и вычисляю тип y
	using TypeY = typename std::remove_cv< typename std::remove_reference<decltype(std::invoke(proj_y, left))>::type >::type;
	using X = typename std::remove_cv<typename std::remove_reference<decltype(std::invoke(proj_x, left))>::type >::type;
	static_assert( std::is_same_v<X, TypeX> , "ERROR something wrong for x type in linear_interpol");
	const TypeY & y1 = std::invoke(proj_y, left);
	const TypeY & y2 = std::invoke(proj_y, right);
	const TypeX & x1  = std::invoke(proj_x, left);
	const TypeX & x2  = std::invoke(proj_x, right);
	double p;
	if constexpr ( std::is_same_v < TypeX, std::chrono::system_clock::time_point> ) {
		//Wowrkaround special case when directo cast to double is impossible
		p = static_cast<double>((x-x1).count()) / static_cast<double>((x2-x1).count());
	} 
	else {
		p = static_cast<double>(x-x1) / static_cast<double>(x2-x1);
	}
	return y1 +  p*(y2-y1);
}

DepolarizerState Depolarizer::get_state(void) const 
{
	return {
		std::chrono::system_clock::now(),
		isscan,
		ison,
		harmonic_number,
		frequency,
		speed,
		step,
		width,
		attenuation
	};
}

/*
std::optional<Depolarizer::DepolarizerState> Depolarizer::get_state(const std::chrono::system_clock::time_point & t) const 
{
	auto begin_search_timestamp = std::chrono::steady_clock::now();
	//fmt::print(" get_state {:.0f}   \n", t.time_since_epoch().count()*1e-9);
	if( flist.empty() ) return std::nullopt; 
	auto first = flist.front();
	auto last = last_frequency_list_point_iterator;
	//fmt::print(" диапазон {:.0f}  {:.0f}  \n", first.timestamp.time_since_epoch().count()*1e-9, last->timestamp.time_since_epoch().count()*1e-9);
	//fmt::print(" частоты  {:.3f}  {:.3f}  \n", first.frequency, last->frequency);
	auto left  = flist.before_begin();
	auto right = flist.end();
	//пробегаем по списку от начала и прерываемся, как только найдем момент времени
	//больше искомого. Этот момент времени будет правой границей (right). 
	//При этом предыдущее значение в списке будет записано в левую границу (left)
	for ( auto it = flist.begin(); it!=flist.end(); ++it) {
		if ( it->timestamp > t) {
			right = it;
			break;
		}
		//приходится каждый раз запоминать предыдущее, т.к. у forward_list нет ссылки на
		//предыдущий элемент.
		left=it;
	}

	if ( right ==  flist.begin() ) {
		//Если самое первое значение уже больше искомого, то определить время
		//мы не можем, g
		//fmt::print("{:.0f}  {:%F %T} самое первое значение больше искомого \n", t.time_since_epoch().count()*1e-9, t);
		return std::nullopt;
	} 

	if ( right == flist.end() ) {
		//если дошли до конца списка или список пуст и время большее искомого не нашли,
		//то имеется два случая
		if ( left->timestamp == t) {
			//последняя запись в списке имеет искомое время
			//fmt::print("range ( {:%F %T}, ... )   frequency = {}\n", left->t, left->frequency);
			return *left;
		} 
		//вообще нет никакого времени
		fmt::print("{:.0f}  {:%F %T} вообще нет никакого времени\n", t.time_since_epoch().count()*1e-9, t);
		return std::nullopt;
		//return { right->t, right->frequency };
	} 
	auto end_search_timestamp = std::chrono::steady_clock::now();
	fmt::print("Время поиска записи: {} ms",  (end_search_timestamp - begin_search_timestamp).count()*1e-6 );
	//linear interpolation
	//auto li = [&left, &right, &t] ( auto proj ) {
	//	return linear_interpol(proj, &DepolarizerState::timestamp,  *left, *right, t);
	//};
	auto t1 = left->timestamp;
	auto t2 = right->timestamp;
	double f1 = left->frequency;
	double f2 = right->frequency;

	//линейной интерполяцией находим промежуточную частоту.
	//double frequency = f1 + (f2-f1)*static_cast<double>((t-t1).count())/static_cast<double>((t2-t1).count());
	//хотя это не правильно, ведь в данный момент генерируется именно левая (боллее раняя частота). 
	//надо проверить что это так.
	//fmt::print(" left.freq = {} right.freq = {}\n" , f1, f2);
	// Проверил так и есть.
	// поэтому буду брать левую частоту

	//fmt::print("{} fr = {} Hz, speed = {} Hz/s,  att = {} dB", li(&DepolarizerState::frequency), li(&DepolarizerState::speed), li(&DepolarizerState::attenuation));
	return DepolarizerState { 
		t,
		left->isscan,
		left->ison,
		left->harmonic_number,
		//li(&DepolarizerState::frequency),
		left->frequency,
		left->speed,
		left->step,
		left->width,
		left->attenuation
	};
}
*/

std::optional<Depolarizer::DepolarizerState> Depolarizer::get_state(const std::chrono::system_clock::time_point & t) const 
{
	auto begin_search_timestamp = std::chrono::steady_clock::now();
	//fmt::print(" get_state {:.0f}   \n", t.time_since_epoch().count()*1e-9);
	if( flist.empty() ) return std::nullopt; 
	//Будем осуществляеть поиск в обратном направлении, чтобы быстрее находить
	//послдение данные
	auto left  = flist.rend();
	auto right = flist.rbegin(); //смотрим на последний во времени элемент
	for ( auto it = flist.rbegin(); it != flist.rend(); ++it) {
		if ( it->timestamp <= t) {
			left = it; //нашли элемент время у которого меньше либо равно искомому
			break; //тут же прервемся
		}
		right = it;
	}

	//самый последний элемент имеет время
	//мненьше искомого. Тут мы ничего не можем
	//гарантировать поэтому возвращаем консервативный ответ
	if ( left ==  flist.rbegin() ) {
		if ( (t - left->timestamp) < step_time )  return *left;
		return std::nullopt;
	} 

	//дошли до начала списка и элементов  меньшим временем не существует
	if ( left == flist.rend() ) {
		return std::nullopt;
	} 

	// в этом месте у нас должна быть найдена пара значений
	auto end_search_timestamp = std::chrono::steady_clock::now();
	fmt::print("Время поиска записи: {} ms\n",  (end_search_timestamp - begin_search_timestamp).count()*1e-6 );
	//linear interpolation
	//auto li = [&left, &right, &t] ( auto proj ) {
	//	return linear_interpol(proj, &DepolarizerState::timestamp,  *left, *right, t);
	//};
	auto t1 = left->timestamp;
	auto t2 = right->timestamp;
	double f1 = left->frequency;
	double f2 = right->frequency;
	fmt::print(" left.freq      = {:>34}     right.freq = {:>34}\n" , f1, f2);
	fmt::print(" left.timestamp = {:>34.3f} right.timetamp = {:>34.3f}\n" , t1.time_since_epoch().count()*1e-9, t2.time_since_epoch().count()*1e-9);
	//double frequency = f1 + (f2-f1)*static_cast<double>((t-t1).count())/static_cast<double>((t2-t1).count());
	//линейной интерполяцией находим промежуточную частоту.
	//хотя это не правильно, ведь в данный момент генерируется именно левая (боллее раняя частота). 
	//надо проверить что это так.
	// Проверил так и есть.
	// поэтому буду брать левую частоту
	//fmt::print("{} fr = {} Hz, speed = {} Hz/s,  att = {} dB", li(&DepolarizerState::frequency), li(&DepolarizerState::speed), li(&DepolarizerState::attenuation));
	
	return DepolarizerState { 
		t,
		left->isscan,
		left->ison,
		left->harmonic_number,
		//li(&DepolarizerState::frequency),
		left->frequency,
		left->speed,
		left->step,
		left->width,
		left->attenuation
	};
}
