#ifndef IBN_DEPOLARIZER_READ_CONFIG_H
#define IBN_DEPOLARIZER_READ_CONFIG_H

#include <list>
#include <memory>
#include <tuple>
#include <iostream>

#include <fmt/core.h>
#include <fmt/color.h>
#include <fmt/ostream.h>

#include <boost/property_tree/info_parser.hpp>

#include "Depolarizer.h"

#include "camac/Generator.h"
#include "camac/DelayLine.h"
#include "camac/Attenuator.h"

#include "device/ArrayAttenuator.h"

#include "device/FakeGenerator.h"
#include "device/FakeAttenuator.h"
#include "device/FakeDelayLine.h"

#include "print.h"


	template<typename ... Types>
void print(std::ostream & os, const Types & ... msgs) 
{
	using namespace std::chrono;
	std::string fmt_string="{:%F %T}.{:03}: ";
	constexpr size_t N = sizeof...(msgs);
	if(N==0) return;
	for(size_t i=0; i<N-1; ++i) {
		fmt_string += "{}: ";
	}
	fmt_string += "{}";
	fmt_string += "\n";
	auto now = system_clock::now();
	auto round_now = round<milliseconds>(now);
	auto ms = (round_now.time_since_epoch() % 1s).count();
	fmt::print(os, fmt::runtime(fmt_string), now,ms , msgs...);
	//os << std::flush;
}


inline auto create_attenuator(std::string name, std::string model, std::string addr, std::string shift="") -> device::Attenuator* 
{
	device::Attenuator * attenuator = nullptr;
	if ( model == "ARRAY" ) {
		attenuator = new device::ArrayAttenuator(name);
	} else 
		if( model == "CAMAC/A0608/15" ) {
			attenuator = new camac::Attenuator_A0608_15(camac::Address(addr), name);
		} else 
			if( model == "CAMAC/A0608/45" ) {
				attenuator  = new camac::Attenuator_A0608_45(camac::Address(addr), name);
			} else 
				if ( model == "FAKE" ) {
					attenuator = new device::FakeAttenuator(name, addr);
				}

	if(attenuator) {
		print(std::clog, "INFO", shift+"Init", attenuator->GetInfo(), "atenuation: ", attenuator->GetAttenuation(), "dB");
	}
	return attenuator;
}

struct GeneratorReferenceSignalConfig_t {
	enum Type {
		Internal,
		External
	} type;
	double reference_frequency = 5e6;
};

inline auto create_generator(std::string name, std::string model, std::string addr, GeneratorReferenceSignalConfig_t rs, std::string shift="") -> device::Generator* 
{
	device::Generator * ptr{nullptr};
	int version{0};
	int sn{0};
	if ( model == "CAMAC/DDS500CAM" ) {
		auto g  = new camac::Generator_DDS500CAM(camac::Address(addr), name);
		auto rst = g->GetReferenceSignalType();
		switch ( rs.type ) {
			case GeneratorReferenceSignalConfig_t::Type::Internal:
				//переключаем только если внутрянняя опорная частота не была установлена
				if ( rst != camac::Generator_DDS500CAM::ReferenceSignal::Internal ) {
					g->SetInternalReferenceSignal();
				}
				break;
			case GeneratorReferenceSignalConfig_t::Type::External:
				//переключаем только если внешняя опорная частота не была установлена или сама частота отличается от уставки
				if ( rst != camac::Generator_DDS500CAM::ReferenceSignal::External ||  
					std::abs(g->GetReferenceFrequency()-rs.reference_frequency) <  std::numeric_limits<double>::epsilon()
				   ) 
				{
					g->SetExternalReferenceSignal(1, rs.reference_frequency);
				}
				break;
		}
		version = g->GetVersion();
		sn = g->GetSerialNumber();
		ptr = g;
	} else 
		if ( model == "FAKE" ) {
			ptr  = new device::FakeGenerator(name, addr) ;
		}
	if(ptr) {
		print(std::clog, "INFO", shift+"Init", ptr->GetInfo(), "SN", sn, "version", version);
	}
	return ptr;
}


inline std::list<std::shared_ptr<device::Depolarizer>> get_depolarizer_list(std::string file) 
{
	std::list<std::shared_ptr<device::Depolarizer>> depolarizer_list;
	std::shared_ptr<device::Generator> generator;
	std::shared_ptr<device::Attenuator> attenuator;
	std::shared_ptr<device::DelayLine>  delay_line;

	auto now = [] { return std::chrono::system_clock::now(); };

	print(std::clog, "INFO", "Initializing hardware using configuration in file ", file);
	boost::property_tree::ptree tree;
	boost::property_tree::info_parser::read_info(file, tree);

	auto get_block_info = [](boost::property_tree::ptree & tr) {
		std::string name = tr.data();
		auto model = tr.get_optional<std::string>("model").get_value_or("FAKE");
		auto addr  = tr.get_optional<std::string>("address").get_value_or("-1:-1:-1");
		return std::tuple{name, model, addr};
	};

	try {
		for(auto & [key1, tree1] : tree) {
			if(key1 == "Depolarizer") {

				//fmt::print(std::clog, "{:%F %T}: INFO: Parsing {} config for {}\n", now(), key1, tree1.data());
				print(std::clog, "INFO", "Parsing config for ", key1, tree1.data());
				for(auto & [key2, tree2] : tree1) {
					if(key2 == "Attenuator") {
						std::string type = key2;
						auto [name, model, addr ] = get_block_info(tree2);
						auto ptr = create_attenuator(name,model,addr,"\t\t");
						if(ptr) {
							attenuator.reset(ptr);
							if( model == "ARRAY" ) {
								for( auto & [key3, tree3] : tree2) {
									//std::cout << "key3 = " << key3 << std::endl;
									if(key3=="model") continue;
									auto type = key3;
									auto [name, model, addr ] = get_block_info(tree3);
									auto ptr = create_attenuator(name, model,addr,"\t\t\t\t");
									if(ptr) std::dynamic_pointer_cast<device::ArrayAttenuator>(attenuator)->AddAttenuator(ptr);
								}
							}
						}
					}

					if(key2 == "DelayLine") {
						std::string type = key2;
						auto [name, model, addr ] = get_block_info(tree2);
						if ( model == "CAMAC/L0606" ) {
							auto step = tree2.get_optional<size_t>("step").get_value_or(4);
							auto ptr = new camac::DelayLine_L0606(name, camac::Address(addr), std::chrono::nanoseconds(step));
							delay_line.reset( ptr );
							print(std::clog, "INFO", "\t\tInit", ptr->GetInfo(), ptr->GetDelay());
						}
					}

					if(key2 == "Generator") {
						std::string type = key2;
						auto [name, model, addr ] = get_block_info(tree2);
						std::string ref_freq_type = tree2.get_optional<std::string>("reference_frequency_type").get_value_or("internal");
						GeneratorReferenceSignalConfig_t rs;
						if (ref_freq_type  == "internal" ) rs.type = GeneratorReferenceSignalConfig_t::Type::Internal;
						if (ref_freq_type  == "external" ) rs.type = GeneratorReferenceSignalConfig_t::Type::External;
						rs.reference_frequency = tree2.get_optional<double>("reference_frequency").get_value_or(5e6);
						auto ptr = create_generator(name, model,addr,rs,"\t\t");
						if(ptr) generator.reset(ptr);
						else  generator.reset( create_generator(name,"FAKE",addr,rs, "\t\t") ); 
					}
				}
				if(!generator) std::runtime_error("Unable to find configuration for Generator");
				if(!attenuator)std::runtime_error("Unable to find configuration for Attenuator"); 
				if(!delay_line)std::runtime_error("Unable to find configuration for DelayLine"); 
				depolarizer_list.push_back(std::make_shared<Depolarizer>(tree1.data(), generator, attenuator, delay_line) );
				print(std::clog, "INFO", "Init",  depolarizer_list.back()->GetInfo(), " with ", generator->GetInfo().Name(), attenuator->GetInfo().Name(), delay_line->GetInfo().Name());
			}
		}
	} catch (std::runtime_error & e) {
		//print(std::cerr, "{:%F %T}: ", now());
		print(std::cerr, "ERROR: Unable to init hardware. Problaly camac crates are turned OFF.", e.what());
		//fmt::print(std::clog, "{:%F %T}: INFO: Switch to  FAKE hardware for some or all hardware.\n", now());
		//if (! generator ) generator.reset( new device::FakeGenerator("fake_generator") );
		//if (! attenuator ) attenuator.reset( new device::FakeAttenuator("fake_generator") );
		//if (! delay_line ) delay_line.reset( new device::FakeDelayLine("fake_generator") );

		//if( generator and attenuator and delay_line ) {
		//	depolarizer_list.push_back(std::make_shared<Depolarizer>("fake_depolarizer", generator, attenuator, delay_line) );
		//	fmt::print(std::clog, "{:%F %T}: INFO: Init {}: \"{}\"\n", now(), "Depolarizer", "fake_depolarizer");
		//}
	}
	std::clog << std::flush;
	return depolarizer_list;
}

#endif
