/*
 * =====================================================================================
 *
 *       Filename:  main.cpp
 *
 *    Description:  Run server for the depolarizer device
 *
 *        Version:  1.0
 *        Created:  14.06.2017 20:31:56
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *   Organization:  Budker Insitute of Nuclear Physics
 *
 * =====================================================================================
 */

#include <string>
#include <csignal>
#include <algorithm>
#include <boost/program_options.hpp>
#include "DepolarizerServer.h"
#include "Depolarizer.h"

#include "ReadConfig.h"
#include <ranges>

#include <fmt/core.h>
#include <fmt/xchar.h>


#include "version.h"

std::shared_ptr<device::Depolarizer> DEPOLARIZER;
void sigint_handler(int)
{
  std::clog << "Catch SIGINT signal (CTRL-C). Turn OFF the depolarizer device and exit" << std::endl;
  if(DEPOLARIZER) {
	  DEPOLARIZER->stop_scan();
	  DEPOLARIZER->off();
  }
  exit(0);
}

auto GCC_VERSION  = fmt::format("{}.{}.{}",__GNUC__, __GNUC_MINOR__ , __GNUC_PATCHLEVEL__);

int main(int argc, char ** argv)
{
	std::ios_base::sync_with_stdio(false);
	namespace po=boost::program_options;
	po::variables_map opt; //options
	po::options_description opt_desc("Allowed options");
	std::string camac_config_file;
	opt_desc.add_options()
		("help","Print this help")
		("port,p",po::value<int>()->default_value(9090), "Port number")
		("config,c",po::value<std::string>()->default_value("Depolarizer.cfg"), "File with camac configuration")
		("verbose,v", po::value<int>()->default_value(static_cast<int>(VerboseLevel::WARNING)), "Verbose level")
		("depolarizer,d", po::value<std::string>()->default_value("depolarizer1"), "Depolarizer name")
		("fake", "Create fake depolarizer");
		;
	try
	{
		po::store(po::parse_command_line(argc, argv, opt_desc), opt);
		po::notify(opt);  
		if (opt.count("help"))
		{
			fmt::print("The depolarizer server: version \"{}\",  build date: {}\n g++ version:{}\n AUTHORS: {}\n{}\nPHONE: {}\nCELL PHONE: {}\nEMAIL: {}\n",
					GIT_VERSION,
					BUILD_DATE,
					GCC_VERSION,
					"Ivan Nikolaev",
					"Budker Institute of Nuclear Physics",
					"48-37",
					"+7(913)987-6819",
					"I.B.Nikolaev@inp.nsk.su");
			//fmt::print("{}", opt_desc);
			std::cout << opt_desc << std::endl;
			exit(0);
		}
		fmt::print("The depolarizer server: version \"{}\",  build date: {}, g++ version {}\n", GIT_VERSION, BUILD_DATE, GCC_VERSION);
		using namespace fmt::literals;
		std::string depolarizer_config_file_name = opt["config"].as<std::string>();
		if ( opt.count("fake") ) {
			print(std::clog, "WARNING", "Run in FAKE mode");
			auto   generator  = std::make_shared<device::FakeGenerator>("fake_generator");
			auto   attenuator  = std::make_shared<device::FakeAttenuator>("fake_attenuator");
			auto   delay_line  = std::make_shared<device::FakeDelayLine>("fake_delay_line");
			DEPOLARIZER = std::make_shared<Depolarizer>("fake_depolarizer", generator, attenuator, delay_line) ;
		}
		else {
			auto depolarizer_list = get_depolarizer_list(depolarizer_config_file_name);
			std::string depolarizer_name = opt["depolarizer"].as<std::string>();
			std::list<std::string> depol_name_list;
			for(auto & d: depolarizer_list) depol_name_list.push_back(d->GetInfo().Name());
			//namespace rv = std::ranges::views;
			//auto depol_name_list = depolarizer_list | rv::transform( [] (const auto & d) -> std::string { return d->GetInfo().Name(); } )  ;
			std::ostringstream oss;
			for( const auto  & x : depol_name_list) {
				oss <<  x << ", ";
			}
			print(std::clog,"INFO", fmt::format("Name of depolarizer in the list: {}", oss.str()));
			
			if( depolarizer_list.empty() ) {
				print(std::cerr,"ERROR",fmt::format("No configuration for depolarizer in file \"{}\"", depolarizer_config_file_name));
				exit(1);
			}
			print(std::clog, "INFO", fmt::format("Found {} depolarizers in file \"{}\"", depolarizer_list.size(), depolarizer_config_file_name));

			//Теперь найдем имя нужного деполяризатора в конфиге
			auto it = std::find_if(depolarizer_list.begin(), depolarizer_list.end(), [&depolarizer_name](const auto & depol) 
					{ return depol->GetInfo().Name()==depolarizer_name ;  });
			if ( it == depolarizer_list.end() ) {
				exit(1);
			}
			DEPOLARIZER = *it;
		}

		print(std::clog, "INFO", fmt::format("Use depolarizer \"{}\"", DEPOLARIZER->GetInfo().Name()));
		std::clog << std::flush;
		std::dynamic_pointer_cast<Depolarizer>(DEPOLARIZER)->set_verbose_level(static_cast<VerboseLevel>(opt["verbose"].as<int>()));
		DepolarizerServer S(opt["port"].as<int>(), DEPOLARIZER);
		signal ( SIGINT,  sigint_handler);
		print(std::cout, "INFO", "Waiting connections");
		std::cout << std::flush;
		S.run();
	}
	catch (po::error & e) {
		fmt::print(std::cerr, "{:%F %T}: ERROR reading config file or parsing command line: {} \n", std::chrono::system_clock::now(), e.what());
	}
}
